@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2018 2ndQuadrant Limited
+# Copyright (C) 2011-2020 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -390,6 +390,7 @@ class BackupManager(RemoteStatusMixin):
             backup_info = LocalBackupInfo(
                 self.server,
                 backup_id=datetime.datetime.now().strftime('%Y%m%dT%H%M%S'))
+            backup_info.set_attribute('systemid', self.server.systemid)
             backup_info.save()
             self.backup_cache_add(backup_info)
             output.info(
@@ -412,6 +413,13 @@ class BackupManager(RemoteStatusMixin):
             # Do the backup using the BackupExecutor
             self.executor.backup(backup_info)
 
+            # Create a restore point after a backup
+            target_name = 'barman_%s' % backup_info.backup_id
+            self.server.postgres.create_restore_point(target_name)
+
+            # Free the Postgres connection
+            self.server.postgres.close()
+
             # Compute backup size and fsync it on disk
             self.backup_fsync_and_set_sizes(backup_info)
 
@@ -428,7 +436,7 @@ class BackupManager(RemoteStatusMixin):
             if backup_info:
                 # Use only the first line of exception message
                 # in backup_info error field
-                backup_info.set_attribute("status", "FAILED")
+                backup_info.set_attribute("status", BackupInfo.FAILED)
                 backup_info.set_attribute(
                     "error",
                     "failure %s (%s)" % (
@@ -450,9 +458,6 @@ class BackupManager(RemoteStatusMixin):
                 self.executor.copy_start_time,
                 human_readable_timedelta(
                     datetime.datetime.now() - executor.copy_start_time))
-            # Create a restore point after a backup
-            target_name = 'barman_%s' % backup_info.backup_id
-            self.server.postgres.create_restore_point(target_name)
 
             # If requested, wait for end_wal to be archived
             if wait:
@@ -861,8 +866,8 @@ class BackupManager(RemoteStatusMixin):
                                 'rebuilding the wal database: %s',
                                 fullname)
                 os.fsync(fxlogdb_new.fileno())
-            shutil.move(xlogdb_new, fxlogdb.name)
-            fsync_dir(os.path.dirname(fxlogdb.name))
+        shutil.move(xlogdb_new, fxlogdb.name)
+        fsync_dir(os.path.dirname(fxlogdb.name))
         output.info('Done rebuilding xlogdb for server %s '
                     '(history: %s, backup_labels: %s, wal_file: %s)',
                     self.config.name, history_count, label_count, wal_count)
@@ -956,7 +961,7 @@ class BackupManager(RemoteStatusMixin):
 
                     # Keeps the WAL segment if it is a newer
                     # than the given backup (the first available)
-                    if backup_info:
+                    if backup_info and backup_info.begin_wal is not None:
                         keep |= wal_info.name >= backup_info.begin_wal
 
                     # If the file has to be kept write it in the new xlogdb
@@ -968,8 +973,8 @@ class BackupManager(RemoteStatusMixin):
                         removed.append(wal_info.name)
                 fxlogdb_new.flush()
                 os.fsync(fxlogdb_new.fileno())
-            shutil.move(xlogdb_new, fxlogdb.name)
-            fsync_dir(os.path.dirname(fxlogdb.name))
+        shutil.move(xlogdb_new, fxlogdb.name)
+        fsync_dir(os.path.dirname(fxlogdb.name))
         return removed
 
     def validate_last_backup_maximum_age(self, last_backup_maximum_age):
