@@ -1,4 +1,4 @@
-# Copyright (C) 2018-2019 2ndQuadrant Limited
+# Copyright (C) 2018-2020 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -15,10 +15,12 @@
 # You should have received a copy of the GNU General Public License
 # along with Barman.  If not, see <http://www.gnu.org/licenses/>.
 
+import bz2
 import collections
 import copy
 import datetime
 import errno
+import gzip
 import json
 import logging
 import multiprocessing
@@ -28,15 +30,16 @@ import shutil
 import signal
 import tarfile
 from functools import partial
-from io import BytesIO
+from io import BytesIO, RawIOBase
 from tempfile import NamedTemporaryFile
 
-from barman.backup_executor import ConcurrentBackupStrategy
+from barman.backup_executor import (ConcurrentBackupStrategy,
+                                    ExclusiveBackupStrategy)
 from barman.fs import path_allowed
 from barman.infofile import BackupInfo
 from barman.postgres_plumbing import EXCLUDE_LIST, PGDATA_EXCLUDE_LIST
 from barman.utils import (BarmanEncoder, force_str, human_readable_timedelta,
-                          total_seconds)
+                          pretty_size, total_seconds)
 
 try:
     import boto3
@@ -59,8 +62,28 @@ except ImportError:
     from Queue import Empty as EmptyQueue
 
 
-DEFAULT_CHUNK_SIZE = 20 << 20
+# S3 multipart upload limitations
+# http://docs.aws.amazon.com/AmazonS3/latest/API/mpUploadUploadPart.html
+MAX_CHUNKS_PER_FILE = 10000
+MIN_CHUNK_SIZE = 5 << 20
+
+# S3 permit a maximum of 5TB per file
+# https://docs.aws.amazon.com/AmazonS3/latest/dev/UploadingObjects.html
+# This is a hard limit, while our upload procedure can go over the specified
+# MAX_ARCHIVE_SIZE - so we set a maximum of 1TB per file
+MAX_ARCHIVE_SIZE = 1 << 40
+
 BUFSIZE = 16 * 1024
+LOGGING_FORMAT = "%(asctime)s [%(process)s] %(levelname)s: %(message)s"
+
+
+def configure_logging(config):
+    """
+    Get a nicer output from the Python logging package
+    """
+    verbosity = config.verbose - config.quiet
+    log_level = max(logging.WARNING - verbosity * 10, logging.DEBUG)
+    logging.basicConfig(format=LOGGING_FORMAT, level=log_level)
 
 
 def copyfileobj_pad_truncate(src, dst, length=None):
@@ -141,10 +164,10 @@ class S3TarUploader(object):
     # We use named temporary files, so we can pass them by name to
     # other processes
     _buffer = partial(NamedTemporaryFile, delete=False,
-                      prefix='barman-cloud-', suffix='.part')
+                      prefix='barman-upload-', suffix='.part')
 
     def __init__(self, cloud_interface, key,
-                 compression=None, chunk_size=DEFAULT_CHUNK_SIZE):
+                 compression=None, chunk_size=MIN_CHUNK_SIZE):
         """
         A tar archive that resides on S3
 
@@ -156,12 +179,13 @@ class S3TarUploader(object):
         self.cloud_interface = cloud_interface
         self.key = key
         self.mpu = None
-        self.chunk_size = chunk_size
+        self.chunk_size = max(chunk_size, MIN_CHUNK_SIZE)
         self.buffer = None
-        self.counter = 1
+        self.counter = 0
         tar_mode = 'w|%s' % (compression or '')
         self.tar = TarFileIgnoringTruncate.open(fileobj=self,
                                                 mode=tar_mode)
+        self.size = 0
         self.stats = None
 
     def write(self, buf):
@@ -170,18 +194,19 @@ class S3TarUploader(object):
         if not self.buffer:
             self.buffer = self._buffer()
         self.buffer.write(buf)
+        self.size += len(buf)
 
     def flush(self):
         if not self.mpu:
             self.mpu = self.cloud_interface.create_multipart_upload(self.key)
         self.buffer.flush()
         self.buffer.seek(0, os.SEEK_SET)
+        self.counter += 1
         self.cloud_interface.async_upload_part(
             mpu=self.mpu,
             key=self.key,
             body=self.buffer,
             part_number=self.counter)
-        self.counter += 1
         self.buffer.close()
         self.buffer = None
 
@@ -191,17 +216,21 @@ class S3TarUploader(object):
         self.flush()
         self.cloud_interface.async_complete_multipart_upload(
             mpu=self.mpu,
-            key=self.key)
+            key=self.key,
+            parts_count=self.counter,
+        )
         self.stats = self.cloud_interface.wait_for_multipart_upload(self.key)
 
 
 class S3UploadController(object):
-    def __init__(self, cloud_interface, key_prefix, compression):
+    def __init__(self, cloud_interface, key_prefix, max_archive_size,
+                 compression):
         """
         Create a new controller that upload the backup in S3
 
         :param CloudInterface cloud_interface: cloud interface instance
         :param str|None key_prefix: path inside the bucket
+        :param int max_archive_size: the maximum size of an archive
         :param str|None compression: required compression
         """
 
@@ -209,6 +238,14 @@ class S3UploadController(object):
         if key_prefix and key_prefix[0] == '/':
             key_prefix = key_prefix[1:]
         self.key_prefix = key_prefix
+        if max_archive_size < MAX_ARCHIVE_SIZE:
+            self.max_archive_size = max_archive_size
+        else:
+            logging.warning("max-archive-size too big. Capping it to to %s",
+                            pretty_size(MAX_ARCHIVE_SIZE))
+            self.max_archive_size = MAX_ARCHIVE_SIZE
+        # We aim to a maximum of MAX_CHUNKS_PER_FILE / 2 chinks per file
+        self.chunk_size = 2 * int(max_archive_size / MAX_CHUNKS_PER_FILE)
         self.compression = compression
         self.tar_list = {}
 
@@ -221,17 +258,22 @@ class S3UploadController(object):
         self.copy_end_time = None
         """Copy end time"""
 
-    def _build_dest_name(self, name):
+    def _build_dest_name(self, name, count=0):
         """
-        Get the name suffix
+        Get the destination tar name
+        :param str name: the name prefix
+        :param int count: the part count
         :rtype: str
         """
-        if self.compression == 'gz':
-            return "%s.tar.gz" % name
-        elif self.compression == 'bz2':
-            return "%s.tar.bz2" % name
-        else:
-            return "%s.tar" % name
+        components = [name]
+        if count > 0:
+            components.append("_%04d" % count)
+        components.append(".tar")
+        if self.compression == "gz":
+            components.append(".gz")
+        elif self.compression == "bz2":
+            components.append(".bz2")
+        return "".join(components)
 
     def _get_tar(self, name):
         """
@@ -242,24 +284,38 @@ class S3UploadController(object):
         """
         if name not in self.tar_list or not self.tar_list[name]:
 
-            self.tar_list[name] = S3TarUploader(
+            self.tar_list[name] = [S3TarUploader(
                 cloud_interface=self.cloud_interface,
                 key=os.path.join(self.key_prefix, self._build_dest_name(name)),
-                compression=self.compression
+                compression=self.compression,
+                chunk_size=self.chunk_size,
+            )]
+        # If the current uploading file size is over DEFAULT_MAX_TAR_SIZE
+        # Close the current file and open the next part
+        uploader = self.tar_list[name][-1]
+        if uploader.size > self.max_archive_size:
+            uploader.close()
+            uploader = S3TarUploader(
+                cloud_interface=self.cloud_interface,
+                key=os.path.join(
+                    self.key_prefix,
+                    self._build_dest_name(name, len(self.tar_list[name]))),
+                compression=self.compression,
+                chunk_size=self.chunk_size,
             )
-        return self.tar_list[name].tar
+            self.tar_list[name].append(uploader)
+        return uploader.tar
 
     def upload_directory(self, label, src, dst, exclude=None, include=None):
-        logging.info("S3UploadController.upload_directory(%r, %r, %r)",
-                     label, src, dst)
-        tar = self._get_tar(dst)
+        logging.info("Uploading '%s' directory '%s' as '%s'",
+                     label, src, self._build_dest_name(dst))
         for root, dirs, files in os.walk(src):
             tar_root = os.path.relpath(root, src)
             if not path_allowed(exclude, include,
                                 tar_root, True):
                 continue
             try:
-                tar.add(root, arcname=tar_root, recursive=False)
+                self._get_tar(dst).add(root, arcname=tar_root, recursive=False)
             except EnvironmentError as e:
                 if e.errno == errno.ENOENT:
                     # If a directory disappeared just skip it,
@@ -275,7 +331,8 @@ class S3UploadController(object):
                     continue
                 logging.debug("Uploading %s", tar_item)
                 try:
-                    tar.add(os.path.join(root, item), arcname=tar_item)
+                    self._get_tar(dst).add(os.path.join(root, item),
+                                           arcname=tar_item)
                 except EnvironmentError as e:
                     if e.errno == errno.ENOENT:
                         # If a file disappeared just skip it,
@@ -285,17 +342,17 @@ class S3UploadController(object):
                         raise
 
     def add_file(self, label, src, dst, path, optional=False):
-        logging.info("S3UploadController.add_file(%r, %r, %r, %r, %r)",
-                     label, src, dst, path, optional)
         if optional and not os.path.exists(src):
             return
+        logging.info("Uploading '%s' file from '%s' to '%s' with path '%s'",
+                     label, src, self._build_dest_name(dst), path)
         tar = self._get_tar(dst)
         tar.add(src, arcname=path)
 
     def add_fileobj(self, label, fileobj, dst, path,
                     mode=None, uid=None, gid=None):
-        logging.info("S3UploadController.add_fileobj(%r, %r, %r)",
-                     label, dst, path)
+        logging.info("Uploading '%s' file to '%s' with path '%s'",
+                     label, self._build_dest_name(dst), path)
         tar = self._get_tar(dst)
         tarinfo = tar.tarinfo(path)
         fileobj.seek(0, os.SEEK_END)
@@ -310,12 +367,14 @@ class S3UploadController(object):
         tar.addfile(tarinfo, fileobj)
 
     def close(self):
-        logging.info("S3UploadController.close()")
+        logging.info("Marking all the uploaded archives as 'completed'")
         for name in self.tar_list:
-            tar = self.tar_list[name]
-            if tar:
-                tar.close()
-                self.upload_stats[name] = tar.stats
+            if self.tar_list[name]:
+                # Tho only opened file is the last one, all the others
+                # have been already closed
+                self.tar_list[name][-1].close()
+                self.upload_stats[name] = [tar.stats
+                                           for tar in self.tar_list[name]]
             self.tar_list[name] = None
 
         # Store the end time
@@ -327,7 +386,7 @@ class S3UploadController(object):
 
         :rtype: dict
         """
-        logging.info("S3UploadController.statistics()")
+        logging.info("Calculating backup statistics")
 
         # This method can only run at the end of a non empty copy
         assert self.copy_end_time
@@ -350,26 +409,33 @@ class S3UploadController(object):
         upload_end = None
         serialized_time = datetime.timedelta(0)
         for name in self.upload_stats:
-            data = self.upload_stats[name]
-            logging.debug('Calculating statistics for file %s, data: %s',
-                          name, json.dumps(data, indent=2, sort_keys=True,
-                                           cls=BarmanEncoder))
-            if upload_start is None or upload_start > data['start_time']:
-                upload_start = data['start_time']
-            if upload_end is None or upload_end < data['end_time']:
-                upload_end = data['end_time']
+            name_start = None
+            name_end = None
+            total_time = datetime.timedelta(0)
+            for index, data in enumerate(self.upload_stats[name]):
+                logging.debug(
+                    'Calculating statistics for file %s, index %s, data: %s',
+                    name, index, json.dumps(data, indent=2, sort_keys=True,
+                                            cls=BarmanEncoder))
+                if upload_start is None or upload_start > data['start_time']:
+                    upload_start = data['start_time']
+                if upload_end is None or upload_end < data['end_time']:
+                    upload_end = data['end_time']
+                if name_start is None or name_start > data['start_time']:
+                    name_start = data['start_time']
+                if name_end is None or name_end < data['end_time']:
+                    name_end = data['end_time']
+                parts = data['parts']
+                for num in parts:
+                    part = parts[num]
+                    total_time += part['end_time'] - part['start_time']
+                stat['serialized_copy_time_per_item'][name] = total_seconds(
+                    total_time)
+                serialized_time += total_time
             # Cloud uploads have no analysis
             stat['analysis_time_per_item'][name] = 0
             stat['copy_time_per_item'][name] = total_seconds(
-                data['end_time'] - data['start_time'])
-            parts = data['parts']
-            total_time = datetime.timedelta(0)
-            for num in parts:
-                part = parts[num]
-                total_time += part['end_time'] - part['start_time']
-            stat['serialized_copy_time_per_item'][name] = total_seconds(
-                total_time)
-            serialized_time += total_time
+                name_end - name_start)
 
         # Store the total time spent by copying
         stat['copy_time'] = total_seconds(upload_end - upload_start)
@@ -399,33 +465,53 @@ class FileUploadStatistics(dict):
         part['start_time'] = start_time
 
 
+class StreamingBodyIO(RawIOBase):
+    """
+    Wrap a boto StreamingBody in the IOBase API.
+    """
+    def __init__(self, body):
+        self.body = body
+
+    def readable(self):
+        return True
+
+    def read(self, n=-1):
+        n = None if n < 0 else n
+        return self.body.read(n)
+
+
 class CloudInterface(object):
-    def __init__(self, destination_url, encryption, jobs=2, profile_name=None):
+    def __init__(self, url, encryption, jobs=2,
+                 profile_name=None, endpoint_url=None):
         """
         Create a new S3 interface given the S3 destination url and the profile
         name
 
-        :param str destination_url: Full URL of the cloud destination
+        :param str url: Full URL of the cloud destination/source
         :param str|None encryption: Encryption type string
         :param int jobs: How many sub-processes to use for asynchronous
           uploading, defaults to 2.
         :param str profile_name: Amazon auth profile identifier
+        :param str endpoint_url: override default endpoint detection strategy
+          with this one
         """
-        self.destination_url = destination_url
+        self.url = url
         self.profile_name = profile_name
         self.encryption = encryption
+        self.endpoint_url = endpoint_url
 
         # Extract information from the destination URL
-        parsed_url = urlparse(destination_url)
+        parsed_url = urlparse(url)
         # If netloc is not present, the s3 url is badly formatted.
         if parsed_url.netloc == '' or parsed_url.scheme != 's3':
-            raise ValueError('Invalid s3 URL address: %s' % destination_url)
+            raise ValueError('Invalid s3 URL address: %s' % url)
         self.bucket_name = parsed_url.netloc
-        self.path = parsed_url.path
+        self.bucket_exists = None
+        self.path = parsed_url.path.lstrip('/')
 
         # Build a session, so we can extract the correct resource
         session = boto3.Session(profile_name=profile_name)
-        self.s3 = session.resource('s3')
+        self.s3 = session.resource('s3', endpoint_url=endpoint_url)
 
         # The worker process and the shared queue are created only when
         # needed
@@ -515,6 +601,9 @@ class CloudInterface(object):
             result = self.done_queue.get()
             self.upload_stats[result["key"]].update(result)
 
+        # Raise an error if a job failed
+        self._handle_async_errors()
+
     def _handle_async_errors(self):
         """
         If an upload error has been discovered, stop the upload
@@ -578,7 +667,7 @@ class CloudInterface(object):
         if task["job_type"] == "upload_part":
             if self.abort_requested:
                 logging.info(
-                    "Skipping %s, part %s (worker %s)" % (
+                    "Skipping '%s', part '%s' (worker %s)" % (
                         task["key"],
                         task["part_number"],
                         process_number))
@@ -586,7 +675,7 @@ class CloudInterface(object):
                 return
             else:
                 logging.info(
-                    "Uploading %s, part %s (worker %s)" % (
+                    "Uploading '%s', part '%s' (worker %s)" % (
                         task["key"],
                         task["part_number"],
                         process_number))
@@ -621,7 +710,7 @@ class CloudInterface(object):
                     })
             else:
                 logging.info(
-                    "Completing %s (worker %s)" % (
+                    "Completing '%s' (worker %s)" % (
                         task["key"],
                         process_number))
                 self.complete_multipart_upload(
@@ -642,49 +731,147 @@ class CloudInterface(object):
         Test the S3 connectivity trying to access a bucket
         """
         try:
-            self.s3.Bucket(self.bucket_name).load()
             # We are not even interested in the existence of the bucket,
             # we just want to try if aws is reachable
+            self.bucket_exists = self.check_bucket_existence(self.bucket_name)
             return True
         except EndpointConnectionError as exc:
             logging.error("Can't connect to Amazon AWS/S3: %s", exc)
             return False
 
+    def check_bucket_existence(self, bucket_name):
+        """
+        Search for the target bucket
+        """
+        try:
+            # Search the bucket on s3
+            self.s3.meta.client.head_bucket(Bucket=bucket_name)
+            return True
+        except ClientError as exc:
+            # If a client error is thrown, then check the error code.
+            # If code was 404, then the bucket does not exist
+            error_code = exc.response['Error']['Code']
+            if error_code == '404':
+                return False
+            # Otherwise there is nothing else to do than re-raise the original
+            # exception
+            raise
+
     def setup_bucket(self):
         """
         Search for the target bucket. Create it if not exists
         """
-        try:
-            # Search the bucket on s3
-            self.s3.meta.client.head_bucket(Bucket=self.bucket_name)
-        except ClientError as exc:
-            # If a client error is thrown, then check that it was a 405 error.
-            # If it was a 404 error, then the bucket does not exist.
-            error_code = exc.response['Error']['Code']
-            if error_code == '404':
-                # Get the current region from client.
-                # Do not use session.region_name here because it may be None
-                region = self.s3.meta.client.meta.region_name
-                logging.info(
-                    "Bucket %s does not exist, creating it on region %s",
-                    self.bucket_name, region)
-                create_bucket_config = {
-                    'ACL': 'private',
+        if self.bucket_exists is None:
+            self.bucket_exists = self.check_bucket_existence(self.bucket_name)
+
+        # Create the bucket if it doesn't exist
+        if not self.bucket_exists:
+            # Get the current region from client.
+            # Do not use session.region_name here because it may be None
+            region = self.s3.meta.client.meta.region_name
+            logging.info(
+                "Bucket '%s' does not exist, creating it on region '%s'",
+                self.bucket_name, region)
+            create_bucket_config = {
+                'ACL': 'private',
+            }
+            # The location constraint is required during bucket creation
+            # for all regions outside of us-east-1. This constraint cannot
+            # be specified in us-east-1; specifying it in this region
+            # results in a failure, so we will only
+            # add it if we are deploying outside of us-east-1.
+            # See https://github.com/boto/boto3/issues/125
+            if region != 'us-east-1':
+                create_bucket_config['CreateBucketConfiguration'] = {
+                    'LocationConstraint': region,
                 }
-                # The location constraint is required during bucket creation
-                # for all regions outside of us-east-1. This constraint cannot
-                # be specified in us-east-1; specifying it in this region
-                # results in a failure, so we will only
-                # add it if we are deploying outside of us-east-1.
-                # See https://github.com/boto/boto3/issues/125
-                if region != 'us-east-1':
-                    create_bucket_config['CreateBucketConfiguration'] = {
-                        'LocationConstraint': region,
-                    }
-                self.s3.Bucket(self.bucket_name).create(**create_bucket_config)
+            self.s3.Bucket(self.bucket_name).create(**create_bucket_config)
+            self.bucket_exists = True
+
+    def list_bucket(self, prefix='', delimiter='/'):
+        """
+        List bucket content in a directory manner
+        :param str prefix:
+        :param str delimiter:
+        :return: List of objects and dirs right under the prefix
+        """
+        if prefix.startswith(delimiter):
+            prefix = prefix.lstrip(delimiter)
+
+        res = self.s3.meta.client.list_objects_v2(
+            Bucket=self.bucket_name,
+            Prefix=prefix,
+            Delimiter=delimiter)
+
+        # List "folders"
+        keys = res.get("CommonPrefixes")
+        if keys is not None:
+            for k in keys:
+                yield k.get("Prefix")
+
+        # List "files"
+        objects = res.get("Contents")
+        if objects is not None:
+            for o in objects:
+                yield o.get("Key")
+
+    def download_file(self, key, dest_path, decompress):
+        """
+        Download a file from S3
+
+        :param str key: The S3 key to download
+        :param str dest_path: Where to put the destination file
+        :param bool decompress: Whenever to decompress this file or not
+        """
+        # Open the remote file
+        obj = self.s3.Object(self.bucket_name, key)
+        remote_file = obj.get()['Body']
+
+        # Write the dest file in binary mode
+        with open(dest_path, 'wb') as dest_file:
+            # If the file is not compressed, just copy its content
+            if not decompress:
+                shutil.copyfileobj(remote_file, dest_file)
+                return
+
+            if decompress == 'gzip':
+                source_file = gzip.GzipFile(fileobj=remote_file, mode='rb')
+            elif decompress == 'bzip2':
+                source_file = bz2.BZ2File(remote_file, 'rb')
+            else:
+                raise ValueError("Unknown compression type: %s" % decompress)
+
+            with source_file:
+                shutil.copyfileobj(source_file, dest_file)
+
+    def remote_open(self, key):
+        """
+        Open a remote S3 object and returns a readable stream
+
+        Returns None is the the Key does not exist
+        """
+        try:
+            obj = self.s3.Object(self.bucket_name, key)
+            return StreamingBodyIO(obj.get()['Body'])
+        except ClientError as exc:
+            error_code = exc.response['Error']['Code']
+            if error_code == 'NoSuchKey':
+                return None
             else:
                 raise
 
+    def extract_tar(self, key, dst):
+        """
+        Extract a tar archive from cloud to the local directory
+        """
+        extension = os.path.splitext(key)[-1]
+        compression = '' if extension == '.tar' else extension[1:]
+        tar_mode = 'r|%s' % compression
+        obj = self.s3.Object(self.bucket_name, key)
+        fileobj = obj.get()['Body']
+        with tarfile.open(fileobj=fileobj, mode=tar_mode) as tf:
+            tf.extractall(path=dst)
+
     def upload_fileobj(self, fileobj, key):
         """
         Synchronously upload the content of a file-like object to a cloud key
@@ -771,7 +958,7 @@ class CloudInterface(object):
             'ETag': part['ETag'],
         }
 
-    def async_complete_multipart_upload(self, mpu, key):
+    def async_complete_multipart_upload(self, mpu, key, parts_count):
         """
         Asynchronously finish a certain multipart upload. This method grant
         that the final S3 call will happen after all the already scheduled
@@ -779,6 +966,7 @@ class CloudInterface(object):
 
         :param mpu:  The multipart upload handle
         :param str key: The key to use in the cloud service
+        :param int parts_count: Number of parts
         """
 
         # If an error has already been reported, do nothing
@@ -788,9 +976,12 @@ class CloudInterface(object):
         self._ensure_async()
         self._handle_async_errors()
 
-        # Wait for all the current jobs to be completed and
-        # receive all available updates on worker status
-        self._retrieve_results()
+        # If parts_db has less then expected parts for this upload,
+        # wait for the workers to send the missing metadata
+        while len(self.parts_db[key]) < parts_count:
+            # Wait for all the current jobs to be completed and
+            # receive all available updates on worker status
+            self._retrieve_results()
 
         # Finish the job in S3 to the uploader process
         self.queue.put({
@@ -853,7 +1044,7 @@ class S3BackupUploader(object):
     """
 
     def __init__(self, server_name, postgres, cloud_interface,
-                 compression=None):
+                 max_archive_size, compression=None):
         """
         Object responsible for handling interactions with S3
 
@@ -861,6 +1052,7 @@ class S3BackupUploader(object):
         :param PostgreSQLConnection postgres: The PostgreSQL connection info
         :param CloudInterface cloud_interface: The interface to use to
           upload the backup
+        :param int max_archive_size: the maximum size of an uploading archive
         :param str compression: Compression algorithm to use
         """
 
@@ -868,6 +1060,7 @@ class S3BackupUploader(object):
         self.server_name = server_name
         self.postgres = postgres
         self.cloud_interface = cloud_interface
+        self.max_archive_size = max_archive_size
 
         # Stats
         self.copy_start_time = None
@@ -981,8 +1174,11 @@ class S3BackupUploader(object):
         """
         Upload a Backup  to S3
         """
+        server_name = 'cloud'
         backup_info = BackupInfo(
-            backup_id=datetime.datetime.now().strftime('%Y%m%dT%H%M%S'))
+            backup_id=datetime.datetime.now().strftime('%Y%m%dT%H%M%S'),
+            server_name=server_name,
+        )
         backup_info.set_attribute("systemid", self.postgres.get_systemid())
         key_prefix = os.path.join(
             self.cloud_interface.path,
@@ -991,23 +1187,42 @@ class S3BackupUploader(object):
             backup_info.backup_id
         )
         controller = S3UploadController(
-            self.cloud_interface, key_prefix, self.compression)
-        strategy = ConcurrentBackupStrategy(self.postgres)
-        logging.info("Starting backup %s", backup_info.backup_id)
+            self.cloud_interface,
+            key_prefix,
+            self.max_archive_size,
+            self.compression,
+        )
+        if self.postgres.server_version >= 90600 \
+                or self.postgres.has_pgespresso:
+            strategy = ConcurrentBackupStrategy(self.postgres, server_name)
+        else:
+            strategy = ExclusiveBackupStrategy(self.postgres, server_name)
+        logging.info("Starting backup '%s'", backup_info.backup_id)
         strategy.start_backup(backup_info)
         try:
             self.backup_copy(controller, backup_info)
-            logging.info("Stopping backup %s", backup_info.backup_id)
+            logging.info("Stopping backup '%s'", backup_info.backup_id)
             strategy.stop_backup(backup_info)
-            pgdata_stat = os.stat(backup_info.pgdata)
-            controller.add_fileobj(
-                label='backup_label',
-                fileobj=BytesIO(backup_info.backup_label.encode('UTF-8')),
-                dst='data',
-                path='backup_label',
-                uid=pgdata_stat.st_uid,
-                gid=pgdata_stat.st_gid,
-            )
+
+            # Create a restore point after a backup
+            target_name = 'barman_%s' % backup_info.backup_id
+            self.postgres.create_restore_point(target_name)
+
+            # Free the Postgres connection
+            self.postgres.close()
+
+            # Eventually, add the backup_label from the backup_info
+            if backup_info.backup_label:
+                pgdata_stat = os.stat(backup_info.pgdata)
+                controller.add_fileobj(
+                    label='backup_label',
+                    fileobj=BytesIO(backup_info.backup_label.encode('UTF-8')),
+                    dst='data',
+                    path='backup_label',
+                    uid=pgdata_stat.st_uid,
+                    gid=pgdata_stat.st_gid,
+                )
+
             # Closing the controller will finalize all the running uploads
             controller.close()
 
@@ -1017,6 +1232,9 @@ class S3BackupUploader(object):
             # Store statistics about the copy
             backup_info.set_attribute("copy_stats", controller.statistics())
 
+            # Set the backup status as DONE
+            backup_info.set_attribute("status", BackupInfo.DONE)
+
         # Use BaseException instead of Exception to catch events like
         # KeyboardInterrupt (e.g.: CTRL-C)
         except BaseException as exc:
@@ -1029,7 +1247,7 @@ class S3BackupUploader(object):
                     backup_info.save(file_object=backup_info_file)
                     backup_info_file.seek(0, os.SEEK_SET)
                     key = os.path.join(controller.key_prefix, 'backup.info')
-                    logging.info("Uploading %s", key)
+                    logging.info("Uploading '%s'", key)
                     self.cloud_interface.upload_fileobj(backup_info_file, key)
             except BaseException as exc:
                 # Mark the backup as failed and exit
@@ -1046,9 +1264,6 @@ class S3BackupUploader(object):
             self.copy_start_time,
             human_readable_timedelta(
                 datetime.datetime.now() - self.copy_start_time))
-        # Create a restore point after a backup
-        target_name = 'barman_%s' % backup_info.backup_id
-        self.postgres.create_restore_point(target_name)
 
     def handle_backup_errors(self, action, backup_info, exc):
         """
@@ -1066,9 +1281,147 @@ class S3BackupUploader(object):
         if backup_info:
             # Use only the first line of exception message
             # in backup_info error field
-            backup_info.set_attribute("status", "FAILED")
+            backup_info.set_attribute("status", BackupInfo.FAILED)
             backup_info.set_attribute(
                 "error",
                 "failure %s (%s)" % (action, msg_lines[0]))
         logging.error("Backup failed %s (%s)", action, msg_lines[0])
         logging.debug('Exception details:', exc_info=exc)
+
+
+class BackupFileInfo(object):
+    def __init__(self, oid=None, base=None, path=None, compression=None):
+        self.oid = oid
+        self.base = base
+        self.path = path
+        self.compression = compression
+        self.additional_files = []
+
+
+class S3BackupCatalog(object):
+    """
+    S3 backup catalog
+    """
+    def __init__(self, cloud_interface,
+                 server_name):
+        """
+        Object responsible for retrievin backup catalog from S3
+
+        :param CloudInterface cloud_interface: The interface to use to
+          upload the backup
+        :param str server_name: The name of the server as configured in Barman
+        """
+
+        self.cloud_interface = cloud_interface
+        self.server_name = server_name
+        self.prefix = os.path.join(
+            self.cloud_interface.path,
+            self.server_name,
+            'base')
+        self._backup_list = None
+
+    def get_backup_list(self):
+        """
+        Retrieve the list of available backup from S3
+
+        :rtype: Dict[str,BackupInfo]
+        """
+        if self._backup_list is None:
+
+            backup_list = {}
+
+            # get backups metadata
+            for backup_dir in self.cloud_interface.list_bucket(
+                    self.prefix + '/'):
+                # We want only the directories
+                if backup_dir[-1] != '/':
+                    continue
+                backup_id = os.path.basename(backup_dir.rstrip('/'))
+                backup_info = self.get_backup_info(backup_id)
+
+                if backup_info:
+                    backup_list[backup_id] = backup_info
+            self._backup_list = backup_list
+        return self._backup_list
+
+    def get_backup_info(self, backup_id):
+        """
+        Load a BackupInfo from S3
+
+        :param str backup_id: The backup id to load
+        :rtype: BackupInfo
+        """
+        backup_info_path = os.path.join(self.prefix, backup_id, 'backup.info')
+        backup_info_file = self.cloud_interface.remote_open(backup_info_path)
+        if backup_info_file is None:
+            return None
+        backup_info = BackupInfo(backup_id)
+        backup_info.load(file_object=backup_info_file)
+        return backup_info
+
+    def get_backup_files(self, backup_info):
+        """
+        Get the list of expected files part of a backup
+
+        :param BackupInfo backup_info: the backup information
+        :rtype: dict[int, BackupFileInfo]
+        """
+        # Correctly format the source path on s3
+        source_dir = os.path.join(self.prefix, backup_info.backup_id)
+
+        base_path = os.path.join(source_dir, 'data')
+        backup_files = {
+            None: BackupFileInfo(None, base_path)
+        }
+        if backup_info.tablespaces:
+            for tblspc in backup_info.tablespaces:
+                base_path = os.path.join(source_dir, '%s' % tblspc.oid)
+                backup_files[tblspc.oid] = BackupFileInfo(tblspc.oid,
+                                                          base_path)
+
+        for item in self.cloud_interface.list_bucket(source_dir + '/'):
+            for backup_file in backup_files.values():
+                if item.startswith(backup_file.base):
+                    # Automatically detect additional files
+                    suffix = item[len(backup_file.base):]
+                    # Avoid to match items that are prefix of other items
+                    if not suffix or suffix[0] not in ('.', '_'):
+                        logging.debug("Skipping spurious prefix match: %s|%s",
+                                      backup_file.base, suffix)
+                        continue
+                    # If this file have a suffix starting with `_`,
+                    # it is an additional file and we add it to the main
+                    # BackupFileInfo ...
+                    if suffix[0] == '_':
+                        info = BackupFileInfo(backup_file.oid, base_path)
+                        backup_file.additional_files.append(info)
+                        ext = suffix.split('.', 1)[-1]
+                    # ... otherwise this is the main file
+                    else:
+                        info = backup_file
+                        ext = suffix[1:]
+                    # Infer the compression from the file extension
+                    if ext == 'tar':
+                        info.compression = None
+                    elif ext == 'tar.gz':
+                        info.compression = 'gzip'
+                    elif ext == 'tar.bz2':
+                        info.compression = 'bzip2'
+                    else:
+                        logging.warning("Skipping unknown extension: %s",
+                                        ext)
+                        continue
+                    info.path = item
+                    logging.info(
+                        "Found file from backup '%s' of server '%s': %s",
+                        backup_info.backup_id, self.server_name, info.path)
+                    break
+
+        for backup_file in backup_files.values():
+            if backup_file.path is None:
+                logging.error("Missing file %s.* for server %s",
+                              backup_file.base,
+                              self.server_name)
+                raise SystemExit(1)
+
+        return backup_files
